import re
from functools import lru_cache
from typing import Optional, Set, Dict, Tuple

from django.templatetags.static import static
from django.utils.translation import gettext as _


def bool_icon(b: bool) -> str:
    icon = "admin/img/icon-no.svg"
    if b:
        icon = "admin/img/icon-yes.svg"
    return '<img src="%s" alt="%s">' % (static(icon), b)


class CSPParser:
    fetch_directives = {
        "child-src",
        "connect-src",
        "default-src",
        "font-src",
        "frame-src",
        "img-src",
        "manifest-src",
        "media-src",
        "object-src",
        "prefetch-src",
        "script-src",
        "script-src-elem",
        "script-src-attr",
        "style-src",
        "style-src-elem",
        "style-src-attr",
        "worker-src",
    }
    document_directives = {"base-uri", "plugin-types", "sandbox"}
    navigation_directives = {"form-action", "frame-ancestors", "navigate-to"}
    report_directives = {"report-uri", "report-to"}
    other_directives = {
        "block-all-mixed-content",
        "referrer",
        "require-sri-for",
        "trusted-types",
        "upgrade-insecure-requests",
    }
    all_directives = (
        fetch_directives
        | document_directives
        | navigation_directives
        | report_directives
        | other_directives
    )
    any_sources = {
        "http:",
        "https:",
        "blob:",
        "mediastream:",
        "filesystem:",
        "data:",
        "unsafe-eval",
        "unsafe-hashes",
        "unsafe-inline",
    }

    fetch_source_re = re.compile(
        r"^("
        r"'http:'|'https:'|'ftp:'|'ftps:'|'blob:'|'mediastream:'|'filesystem:'|'data:'|"
        r"http:|https:|ftp:|ftps:|blob:|mediastream:|filesystem:|data:|"
        r"'self'|'unsafe-eval'|'unsafe-hashes'|'unsafe-inline'|'report-sample'|"
        r"'strict-dynamic'|'none'|'nonce-[A-Za-z\d+/]+'|"
        r"'sha256-[A-Za-z\d+/]+'|'sha384-[A-Za-z\d+/]+'|'sha512-[A-Za-z\d+/]+'|"
        r"(http://|https://|ftp://|ftps://)?[\w*\-]+\.[\w.\-]+(:\d{1,5})?"
        r")$"
    )
    navigation_source_re = fetch_source_re
    host_re = re.compile(r"^((http://|https://|ftp://|ftps://)?[\w.\-]+(:\d{1,5})?)$")
    data_re = re.compile(r"^(nonce-|sha256-|sha384-|sha512-)")

    def __init__(self, policies: str, is_secure: bool = True):
        self.policies = policies
        self.is_secure = is_secure
        self.by_directive = {}  # type: Dict[str, Set[str]]

    def load(self):
        if self.policies is None:
            return
        for policy in self.policies.split(";"):
            stripped = policy.strip()
            if not stripped:
                continue
            parsed_policy = self.parse_policy(stripped)
            if parsed_policy:
                directive, sources = parsed_policy
                self.by_directive[directive] = sources

    def parse_policy(self, policy: str) -> Optional[Tuple[str, Set[str]]]:
        directive, sep, arg = policy.partition(" ")
        if directive not in self.all_directives:
            raise ValueError("%s is not recognized" % directive)
        sources = set()
        if directive in self.fetch_directives:
            source_re = self.fetch_source_re
        elif directive in self.navigation_directives:
            source_re = self.navigation_source_re
        elif directive in {"base-uri"}:
            source_re = self.navigation_source_re
        else:
            return
        for source in arg.split(" "):
            if not source or not source_re.match(source):
                continue
            if self.host_re.match(source):
                if source.startswith("http://") or source.startswith("ftp://"):
                    sources.add("http://")
                elif source.startswith("https://") or source.startswith("ftps://"):
                    sources.add("https://")
                else:
                    sources |= {"http://", "https://"}
                continue
            source = source.replace("'", "")
            if self.data_re.match(source):
                sources.add(source.split("-")[0])
                continue
            sources.add(source)
        return directive, sources

    def sources(self, directive: str) -> Set[str]:
        if directive in self.fetch_directives:
            return self.by_directive.get(
                directive, self.by_directive.get("default-src", CSPParser.any_sources),
            )
        return self.by_directive.get(directive, CSPParser.any_sources)

    @property
    def check_unsafe_inline_script(self):
        """Blocks execution of inline JavaScript by not allowing 'unsafe-inline' inside script-src"""
        return "unsafe-inline" not in self.sources("script-src")

    @property
    def check_eval_script(self):
        """Blocks execution of JavaScript's eval() function by not allowing 'unsafe-eval' inside script-src"""
        return "unsafe-eval" not in self.sources("script-src")

    @property
    def check_object(self):
        """Blocks execution of plug-ins, using object-src restrictions"""
        return self.any_sources.isdisjoint(self.sources("object-src"))

    @property
    def check_unsafe_inline_style(self):
        """Blocks inline styles by not allowing 'unsafe-inline' inside style-src"""
        return "unsafe-inline" not in self.sources("style-src")

    @property
    def check_block_active_http_content(self):
        """Blocks loading of active content over HTTP or FTP"""
        active = (
            "connect-src",
            "frame-src",
            "object-src",
            "script-src",
            "script-src-elem",
            "script-src-attr",
            "worker-src",
        )
        return all({"http:"}.isdisjoint(self.sources(x)) for x in active)

    @property
    def check_block_passive_http_content(self):
        """Blocks loading of passive content over HTTP or FTP"""
        passive = (
            "child-src",
            "font-src",
            "img-src",
            "manifest-src",
            "media-src",
            "prefetch-src",
            "style-src",
            "style-src-elem",
            "style-src-attr",
        )
        return all({"http:"}.isdisjoint(self.sources(x)) for x in passive)

    @property
    def check_clickjacking_protection(self):
        """Clickjacking protection, using frame-ancestors"""
        return {"http:", "https:"}.isdisjoint(self.sources("frame-ancestors"))

    @property
    def check_default_deny(self):
        """Deny by default, using default-src 'none'"""
        return self.sources("default-src") == {"none"}

    @property
    def check_base_tag(self):
        """Restricts use of the &lt;base&gt; tag by using base-uri 'none', base-uri 'self', or specific origins"""
        base = self.sources("base-uri")
        return {"none", "self", "http://", "https://"}.issuperset(base)

    @property
    def check_form_destinations(self):
        """Restricts where &lt;form&gt; contents may be submitted by using form-action 'none',
        form-action 'self', or specific URIs"""
        base = self.sources("form-action")
        return {"none", "self", "http://", "https://"}.issuperset(base)

    @property
    def check_script_dynamic(self):
        """Uses CSP3's 'strict-dynamic' directive to allow dynamic script loading (optional)"""
        return "strict-dynamic" in self.sources("script-src")


@lru_cache()
def get_csp_parser(policies: str, is_secure: bool = True) -> Optional[CSPParser]:
    parser = CSPParser(policies, is_secure=is_secure)
    try:
        parser.load()
    except ValueError:
        return None
    return parser


@lru_cache()
def get_csp_analyzis(policies: Optional[str], is_secure: bool = True):
    if policies is None:
        return (
            "<p>%s</p>"
            % _(
                "Content Security Policy (CSP) header not implemented."
                " You should take a look at "
                '<a href="https://django-csp.readthedocs.io/en/latest/">Django CSP</a> to add it.'
            ),
            -25,
        )
    parser = get_csp_parser(policies, is_secure=is_secure)
    if parser is None:
        comment = _(
            "Content Security Policy (CSP) header cannot be parsed successfully."
        )
        score = -25
    elif parser.check_default_deny and all(
        {"unsafe-inline", "unsafe-eval"}.isdisjoint(parser.sources(x))
        for x in parser.fetch_directives
    ):
        comment = _(
            "Content Security Policy (CSP) implemented with default-src 'none'"
            " and without 'unsafe-inline' or 'unsafe-eval'"
        )
        score = 10
    elif all(
        {"unsafe-inline", "unsafe-eval"}.isdisjoint(parser.sources(x))
        for x in parser.fetch_directives
    ):
        comment = _(
            "Content Security Policy (CSP) implemented without 'unsafe-inline' or 'unsafe-eval'"
        )
        score = 5
    elif all(
        {"unsafe-inline", "data:", "http:", "https:"}.isdisjoint(parser.sources(x))
        for x in parser.fetch_directives
        if x != "style-src"
    ):
        comment = _(
            "Content Security Policy (CSP) implemented with unsafe directives inside style-src. "
            "This includes 'unsafe-inline', data:, or overly broad sources such as https: "
        )
        score = 0
    elif any(
        {"http:", "http://"}.intersection(parser.sources(x))
        for x in ("frame-src", "object-src", "media-src", "img-src",)
    ):
        comment = _(
            "Content Security Policy (CSP) implemented, but secure site allows images "
            "or media to be loaded over http"
        )
        score = -10
    elif any("unsafe-eval" in parser.sources(x) for x in parser.fetch_directives):
        comment = _(
            "Content Security Policy (CSP) implemented, but allows 'unsafe-eval'"
        )
        score = -10
    elif parser.is_secure and any(
        {"http://", "http:"}.intersection(parser.sources(x))
        for x in parser.fetch_directives
    ):
        comment = _(
            "Content Security Policy (CSP) implemented, but secure site allows resources to be loaded from http"
        )
        score = -20
    elif {"unsafe-inline", "data:", "https:"}.intersection(
        parser.sources("script-src")
    ) or {"https:"}.intersection(parser.sources("object-src")):
        comment = _(
            "Content Security Policy (CSP) implemented unsafely. This includes 'unsafe-inline' or data:"
            " inside script-src, overly broad sources such as https: inside object-src or script-src, or"
            " not restricting the sources for object-src or script-src."
        )
        score = -20
    elif not parser.is_secure:
        comment = _("Content Security Policy (CSP) implemented, but site is not secure")
        score = -10
    else:
        comment = None
        score = 0
    desc = ""
    if comment:
        desc += "<p>%s</p>\n" % comment
    if parser:
        desc += "<p>%s</p>" % _("HTTPS is assumed during this analyzis.")
        desc += "<table><thead>\n"
        desc += "<tr><th>%s</th><th>%s</th></tr>\n" % (_("Test"), _("Pass"))
        desc += "</thead><tbody>\n"
        for k in (
            "check_unsafe_inline_script",
            "check_eval_script",
            "check_object",
            "check_unsafe_inline_style",
            "check_block_active_http_content",
            "check_block_passive_http_content",
            "check_clickjacking_protection",
            "check_default_deny",
            "check_base_tag",
            "check_form_destinations",
            "check_script_dynamic",
        ):
            desc += "<tr><td>%s</td><td>%s</td></tr>\n" % (
                getattr(CSPParser, k).__doc__,
                bool_icon(getattr(parser, k)),
            )
        desc += "</tbody></table>\n"
    return desc, score



# ##############################################################################
#  This file is part of djt_csp                                                #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import datetime
from functools import lru_cache, cached_property
from html.parser import HTMLParser
from http.cookies import SimpleCookie
from logging import INFO, WARNING, CRITICAL, ERROR
from typing import Dict, List

import pkg_resources
from debug_toolbar.panels import Panel
from django.http import HttpResponse, HttpRequest

# noinspection PyProtectedMember
from django.template import engines, TemplateSyntaxError
from django.templatetags.static import static
from django.utils.safestring import mark_safe

from djt_csp.checkers import (
    CHECKED_RESOURCES,
    CHECKED_HTTP_HEADERS,
    Checker,
    ContentTypeSniff,
    RefererComponent,
    XSSProtection,
    FrameOptions,
    ScriptIntegrityChecker,
    CookieAnalyzer,
    CSPChecker,
    CORSChecker,
    HSTSChecker,
    HPKPChecker,
)


def template_from_string(template_string):
    """
    Convert a string into a template object,
    using a given template engine or using the default backends
    from settings.TEMPLATES if no engine was specified.
    """
    # This function is based on django.template.loader.get_template,
    for engine in engines.all():
        try:
            return engine.from_string(template_string)
        except TemplateSyntaxError:
            pass
    raise TemplateSyntaxError(template_string)


class HeaderHTMLParser(HTMLParser):
    def __init__(self, *, convert_charrefs=True):
        super().__init__(convert_charrefs=convert_charrefs)
        self.headers = {}  # type: Dict[str, str]
        self.scripts_attributes = []  # type: List[Dict[str, str]]
        self.http_ws_resources = {
            x: [] for x in CHECKED_RESOURCES
        }  # type: Dict[str, List[str]]

    def reset(self):
        super().reset()
        self.headers = {}
        self.scripts_attributes = []
        self.http_ws_resources = {x: [] for x in CHECKED_RESOURCES}

    def handle_starttag(self, tag, attrs):
        tag = tag.lower()
        attrs = {x.lower(): y for (x, y) in attrs}
        http_equiv = attrs.get("http-equiv", "").lower()
        content = attrs.get("content")
        if tag == "meta" and http_equiv in CHECKED_HTTP_HEADERS and content:
            self.headers[http_equiv] = content
        elif tag == "script":
            self.scripts_attributes.append(attrs)
        attr = CHECKED_RESOURCES.get(tag)
        origin = attrs.get(tag, "")
        if attr and (origin.startswith("http://") or origin.startswith("ws://")):
            self.http_ws_resources[tag].append(origin)

    def error(self, message):
        pass


class SecurityPanel(Panel):
    """
    """

    name = "Security score"
    title = "Security score"
    template = "templates/debug/security_panel.html"
    has_content = True

    def generate_stats(self, request: HttpRequest, response: HttpResponse):
        content_type = ""
        if response.has_header("Content-Type"):
            content_type = response["Content-Type"]
        if (
            not content_type.startswith("text/html")
            or not isinstance(response, HttpResponse)
            or response.status_code < 200
            or (300 <= response.status_code < 400)
        ):
            return
        max_age = 365 * 24 * 60 * 60  # one year
        expires = datetime.datetime.strftime(
            datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age),
            "%a, %d-%b-%Y %H:%M:%S GMT",
        )
        response_headers = {}
        for header in CHECKED_HTTP_HEADERS:
            if header in response:
                response_headers[header] = response[header]
        parser = HeaderHTMLParser()
        try:
            parser.feed(response.content.decode())
            scripts_attributes = parser.scripts_attributes
            html_headers = parser.headers
            http_ws_resources = parser.http_ws_resources
            cookies_l = []
            cookies = response.cookies  # type: SimpleCookie
            for name, values in cookies.items():
                cookies_l.append(
                    {
                        "name": name,
                        "expires": values["expires"],
                        "path": values["path"],
                        "comment": values["comment"],
                        "domain": values["domain"],
                        "max-age": values["max-age"],
                        "secure": values["secure"],
                        "version": values["version"],
                        "httponly": values["httponly"],
                        "samesite": values["samesite"],
                    }
                )
        except AssertionError:
            scripts_attributes = None
            html_headers = None
            http_ws_resources = None
            cookies_l = None
        values = {
            "cookies": cookies_l,
            "html_headers": html_headers,
            "response_headers": response_headers,
            "content_type": content_type,
            "scripts_attributes": scripts_attributes,
            "http_ws_resources": http_ws_resources,
            "is_secure": request.is_secure(),
        }
        self.record_stats(values)

    @cached_property
    def components(self) -> List[Checker]:
        stats = self.get_stats()
        components = []  # type: List[Checker]
        if stats and stats.get("html_headers") is not None:
            components = [
                CSPChecker(stats),
                CookieAnalyzer(stats),
                CORSChecker(stats),
                HPKPChecker(stats),
                HSTSChecker(stats),
                RefererComponent(stats),
                ScriptIntegrityChecker(stats),
                ContentTypeSniff(stats),
                FrameOptions(stats),
                XSSProtection(stats),
            ]
        for comp in components:
            comp.load()
        return components

    def nav_subtitle(self):
        score = 100
        level = INFO
        for comp in self.components:
            level = max(comp.level, level)
            score += comp.score
        scores = [
            (100, "A+"),
            (90, "A"),
            (85, "A-"),
            (80, "B+"),
            (70, "B"),
            (65, "B-"),
            (60, "C+"),
            (50, "C"),
            (45, "C-"),
            (40, "D+"),
            (30, "D"),
            (25, "D-"),
        ]
        letter = "F"
        for score_, letter_ in scores:
            if score >= score_:
                letter = letter_
                break
        if level >= CRITICAL or score <= 44:
            img = static("admin/img/icon-no.svg")
        elif level >= ERROR or score <= 64:
            img = static("admin/img/icon-no.svg")
        elif level >= WARNING or score <= 84:
            img = static("admin/img/icon-alert.svg")
        else:
            img = static("admin/img/icon-yes.svg")
        return mark_safe(
            "<img src='%s' alt='error'> Grade: %s (%s/100)" % (img, letter, score)
        )

    @property
    def content(self):
        """
        Content of the panel when it's displayed in full screen.

        By default this renders the template defined by :attr:`template`.
        Statistics stored with :meth:`record_stats` are available in the
        template's context.
        """
        template = self.get_template()
        context = {"components": self.components}
        content = template.render(context)
        return content

    @lru_cache()
    def get_template(self):
        template_filename = pkg_resources.resource_filename("djt_csp", self.template)
        with open(template_filename) as fd:
            template_content = fd.read()
        template = template_from_string(template_content)
        return template
